from pathlib import Path
import numpy as np
import pandas as pd
import pyrsm as rsm
import io
from pathlib import Path
from contextlib import redirect_stdout
from shiny import App, render, ui, reactive, Inputs, Outputs, Session
import black
import pyperclip

## next steps

## add a close button (navbar?) and return something (code?) on close
# https://shiny.rstudio.com/py/api/Session.html#shiny.Session
## adjust plot window height
## render table https://shiny.rstudio.com/py/api/ui.output_table.html#shiny.ui.output_table
## shown description as markdown https://shinylive.io/py/examples/#extra-packages


class model_logit:
    def __init__(self, dataset: pd.DataFrame) -> None:
        self.dataset = dataset

    def shiny_ui(self):
        return ui.page_fluid(
            ui.head_content(
                ui.tags.script((Path(__file__).parent / "www/copy.js").read_text()),
                ui.tags.style((Path(__file__).parent / "www/style.css").read_text()),
                # ui.tags.script("copy.js"), # not getting picked up for some reason
                # ui.tags.style("style.css"), # not getting picked up for some reason
                # ui.tags.script(
                #     src="https://mathjax.rstudio.com/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML"
                # ),
                # ui.tags.script(
                #     "if (window.MathJax) MathJax.Hub.Queue(['Typeset', MathJax.Hub]);"
                # ),
            ),
            ui.layout_sidebar(
                ui.panel_sidebar(
                    ui.panel_well(
                        ui.input_file(
                            id="logistic_data",
                            label="Upload a Pickle file",
                            accept=[".pkl"],
                            multiple=False,
                        ),
                    ),
                    ui.panel_conditional(
                        "input.tabs_logistic == 'Summary'",
                        ui.panel_well(
                            ui.input_action_button(
                                "run",
                                "Estimate model",
                                class_="btn-success",
                                width="100%",
                            ),
                        ),
                        ui.panel_well(
                            ui.output_ui("ui_rvar"),
                            ui.output_ui("ui_evar"),
                            width=3,
                        ),
                    ),
                    ui.panel_conditional(
                        "input.tabs_logistic == 'Plot'",
                        ui.panel_well(
                            ui.input_action_button(
                                "plot",
                                "Create plot",
                                class_="btn-success",
                                width="100%",
                            ),
                        ),
                        ui.panel_well(
                            ui.input_select(
                                id="logistic_plots",
                                label="Plots",
                                selected=None,
                                choices={
                                    "or": "OR plot",
                                    "pred": "Prediction plot",
                                    "vimp": "Permutation importance",
                                },
                            ),
                            width=3,
                        ),
                    ),
                ),
                ui.panel_main(
                    ui.navset_tab_card(
                        ui.nav(
                            "Data",
                            ui.output_ui("show_data"),
                            ui.output_ui("show_data_code"),
                        ),
                        ui.nav(
                            "Summary",
                            ui.output_text_verbatim("logistic_summary"),
                            ui.output_ui("show_summary_code"),
                        ),
                        ui.nav(
                            "Plot",
                            ui.output_plot("logistic_plot"),
                            ui.output_ui("show_plot_code"),
                        ),
                        id="tabs_logistic",
                    )
                ),
            ),
        )

    def shiny_server(self, input: Inputs, output: Outputs, session: Session):
        @reactive.Calc
        def load_data():
            if input.logistic_data() is None:
                # if not "mydata" in globals():
                #     return "Please upload a Pickle file", "", ""
                # else:
                #     data = globals()["mydata"]
                #     return data, "data", "data = pd.read_pickle('data.pkl')"
                return self.dataset, "data", "data = pd.read_pickle('data.pkl')"
            else:
                f = input.logistic_data()
                file_name = f[0]["name"]
                fname = file_name.split(".")[0]
                code = f"""{fname} = pd.read_pickle("{file_name}")"""
                return pd.read_pickle(f[0]["datapath"]), fname, code

        @output(id="show_data")
        @render.ui
        def show_data():
            if input.logistic_data() is None:
                return "Please upload a Pickle file"
            else:
                data, _, _ = load_data()
                return ui.HTML(
                    data.head().to_html(
                        classes="table table-striped data_preview", index=False
                    )
                )

        @output(id="show_data_code")
        @render.ui
        def show_data_code():
            if input.logistic_data() is None:
                return ""
            else:
                _, _, code = load_data()
                return code_formatter(code)

        @output(id="ui_rvar")
        @render.ui
        def ui_rvar():
            data, _, _ = load_data()
            if isinstance(data, str):
                df_cols = []
            else:
                df_cols = list(data.columns)
            return ui.input_select(
                id="rvar",
                label="Response Variable",
                selected=None,
                choices=df_cols,
            )

        @output(id="ui_evar")
        @render.ui
        def ui_evar():
            data, _, _ = load_data()
            if isinstance(data, str):
                df_cols = []
            else:
                df_cols = list(data.columns)
                if (input.rvar() is not None) and (input.rvar() in df_cols):
                    df_cols.remove(input.rvar())
            return ui.input_select(
                id="evar",
                label="Explanatory Variables",
                selected=None,
                choices=df_cols,
                multiple=True,
                selectize=False,
            )

        @reactive.Calc
        def logistic_regression():
            data, _, _ = load_data()
            return rsm.logistic(
                dataset=data, rvar=input.rvar(), evar=list(input.evar())
            )

        @output(id="logistic_summary")
        @render.text
        @reactive.event(input.run, ignore_none=True)
        def logistic_summary():
            out = io.StringIO()
            with redirect_stdout(out):
                logistic_regression().summary()
            return out.getvalue()

        def code_formatter(code):
            cmd = black.format_str(code, mode=black.Mode())
            pyperclip.copy(cmd)
            cmd = f"""<pre><details><summary>Code</summary>{cmd}</details></pre>"""
            return ui.HTML(cmd)

        @output(id="show_summary_code")
        @render.text
        @reactive.event(input.run, ignore_none=True)
        def show_summary_code():
            rvar = input.rvar()
            evar = list(input.evar())
            _, fname, code = load_data()
            cmd = f"""{code}\nlr = rsm.logistic(dataset={fname}, rvar="{rvar}", evar={evar})\nlr.summary()"""
            return code_formatter(cmd)

        @output(id="logistic_plot")
        @render.plot()
        @reactive.event(input.plot, ignore_none=True)
        def logistic_plot():
            return logistic_regression().plot(plots=input.logistic_plots())

        @output(id="show_plot_code")
        @render.text
        @reactive.event(input.plot, ignore_none=True)
        def show_plot_code():
            rvar = input.rvar()
            evar = list(input.evar())
            plot_type = input.logistic_plots()
            _, fname, code = load_data()
            cmd = f"""{code}\nlr = rsm.logistic(dataset={fname}, rvar="{rvar}", evar={evar})\nlr.plot("{plot_type}")"""
            return code_formatter(cmd)


## should work based on https://shinylive.io/py/examples/#static-content
## but doesn't
# www_dir = Path(__file__).parent / "www"
# app = App(app_ui, server, static_assets=www_dir)


## uncomment for development and testing
# rsm.load_data(pkg="data", name="titanic", dct=globals())
# sl = shiny_logit(titanic)
# app = App(sl.shiny_ui(), sl.shiny_server)

## uncomment for development and testing
# rsm.load_data(pkg="data", name="titanic", dct=globals())


# def launch_logit(data):
#     lr = model_logit(data)
#     nest_asyncio.apply()
#     webbrowser.open("http://127.0.0.1:8000")
#     uvicorn.run(App(lr.shiny_ui(), lr.shiny_server), port=8000)


# launch_logit(titanic)



import pandas as pd
import statsmodels.formula.api as smf
from typing import Optional
from statsmodels.regression.linear_model import RegressionResults as rrs
from shiny import App
from .radiant.regress import model_regress
from .utils import ifelse, format_nr, setdiff
from .visualize import pred_plot_sm, vimp_plot_sm
from .model import (
    sig_stars,
    model_fit,
    extract_evars,
    extract_rvar,
    scatter_plot,
    reg_dashboard,
    residual_plot,
    coef_plot,
    coef_ci,
    predict_ci,
)
from .model import vif as calc_vif
from .visualize import distr_plot
from .basics import correlation


class regress:
    def __init__(
        self,
        dataset: pd.DataFrame,
        rvar: Optional[str] = None,
        evar: Optional[list[str]] = None,
        int: Optional[list[str]] = None,
        form: Optional[str] = None,
    ) -> None:

        """
        Estimate linear regression model

        Parameters
        ----------
        dataset: pandas DataFrame; dataset
        evar: List of strings; contains the names of the columns of data to be used as explanatory variables
        rvar: String; name of the column to be used as the response variable
        form: String; formula for the regression equation to use if evar and rvar are not provided
        """
        self.dataset = dataset
        self.rvar = rvar
        self.evar = ifelse(isinstance(evar, str), [evar], evar)
        self.int = ifelse(isinstance(int, str), [int], int)
        self.form = form

        if self.form:
            self.fitted = smf.ols(formula=self.form, data=self.dataset).fit()
            self.evar = extract_evars(self.fitted.model, self.dataset.columns)
            self.rvar = extract_rvar(self.fitted.model, self.dataset.columns)
        else:
            self.form = f"{self.rvar} ~ {' + '.join(self.evar)}"
            if self.int:
                self.form += f"+ {' + '.join(self.int)}"
            self.fitted = smf.ols(self.form, data=self.dataset).fit()

        df = pd.DataFrame(self.fitted.params, columns=["coefficient"]).dropna()
        df["std.error"] = self.fitted.params / self.fitted.tvalues
        df["t.value"] = self.fitted.tvalues
        df["p.value"] = self.fitted.pvalues
        df["  "] = sig_stars(self.fitted.pvalues)
        self.coef = df.reset_index()

    def summary(
        self, ssq=False, vif=False, ci=False, dec=3, name="Not provided", shiny=False
    ) -> None:
        """
        Summarize output from a linear regression model

        parameters
        ----------
        ssq: Boolean; if True, include sum of squares
        vif: Boolean; if True, include variance inflation factors
        """
        print("Linear regression (OLS)")
        print("Data                 :", name)
        print("Response variable    :", self.rvar)
        print("Explanatory variables:", ", ".join(self.evar))
        print(f"Null hyp.: the effect of x on {self.rvar} is zero")
        print(f"Alt. hyp.: the effect of x on {self.rvar} is not zero")

        df = self.coef.copy()
        df["coefficient"] = df["coefficient"].round(2)
        df["std.error"] = df["std.error"].round(dec)
        df["t.value"] = df["t.value"].round(dec)
        df["p.value"] = ifelse(
            df["p.value"] < 0.001, "< .001", df["p.value"].round(dec)
        )
        df = df.set_index("index")
        df.index.name = None
        print(f"\n{df.to_string()}")
        print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")
        print(f"\n{model_fit(self.fitted, shiny=shiny)}")

        if ci:
            print("\nConfidence intervals:")
            print(f"\n{coef_ci(self.fitted).to_string()}")

        if ssq:
            print("\nSum of squares:")
            index = ["Regression", "Error", "Total"]
            sum_of_squares = [
                self.fitted.ess,
                self.fitted.ssr,
                self.fitted.centered_tss,
            ]
            sum_of_squares = pd.DataFrame(index=index).assign(
                df=format_nr(
                    [
                        self.fitted.df_model,
                        self.fitted.df_resid,
                        self.fitted.df_model + self.fitted.df_resid,
                    ],
                    dec=0,
                ),
                SS=format_nr(sum_of_squares, dec=0),
            )
            print(f"\n{sum_of_squares.to_string()}")

        if vif:
            print("\nVariance inflation factors:")
            print(f"\n{calc_vif(self.fitted).to_string()}")

    def predict(self, df=None, ci=False, alpha=0.05) -> pd.DataFrame:
        """
        Predict values for a linear regression model
        """
        if df is None:
            df = self.dataset
        df = df.loc[:, self.evar].copy()
        if ci:
            return pd.concat([df, predict_ci(self.fitted, df, alpha=alpha)], axis=1)
        else:
            pred = pd.DataFrame().assign(prediction=self.fitted.predict(df))
            return pd.concat([df, pred], axis=1)

    def plot(
        self,
        plots="dist",
        nobs: int = 1000,
        intercept=False,
        alpha=0.05,
        incl=None,
        excl=[],
        incl_int=[],
        fix=True,
        hline=False,
        figsize=None,
    ) -> None:
        """
        Plots for a linear regression model
        """
        dataset = self.dataset[[self.rvar] + self.evar].copy()
        if "dist" in plots:
            distr_plot(dataset)
        if "corr" in plots:
            cr = correlation(dataset)
            cr.plot(nobs=nobs, figsize=figsize)
        if "scatter" in plots:
            scatter_plot(self.fitted, dataset, nobs=nobs, figsize=figsize)
        if "dashboard" in plots:
            reg_dashboard(self.fitted, nobs=nobs)
        if "residual" in plots:
            residual_plot(self.fitted, dataset, nobs=nobs, figsize=figsize)
        if "pred" in plots:
            pred_plot_sm(
                self.fitted,
                self.dataset,
                incl=incl,
                excl=excl,
                incl_int=incl_int,
                fix=fix,
                hline=hline,
                nnv=20,
                minq=0.025,
                maxq=0.975,
            )
        if "vimp" in plots:
            vimp_plot_sm(self.fitted, self.dataset, rep=10, ax=None, ret=False)
        if "coef" in plots:
            coef_plot(
                self.fitted,
                alpha=alpha,
                intercept=intercept,
                incl=incl,
                excl=excl,
                figsize=figsize,
            )

    def f_test(self, vtt=None, dec=3) -> None:
        """
        F-test for competing models

        Parameters
        ----------
        vtt : list
            List of strings; contains the names of the columns of data to be tested
        """
        if vtt is None:
            form = f"{self.rvar} ~ 1"
            vtt = self.evar.copy()
        else:
            evar = setdiff(self.evar, vtt)
            if len(evar) == 0:
                form = f"{self.rvar} ~ 1"
            else:
                form = f"{self.rvar} ~ {' + '.join(evar)}"

        vtt = ifelse(isinstance(vtt, str), [vtt], vtt)
        hypothesis = [f"({v} = 0)" for v in vtt]
        print(f"Model 1: {form}")
        print(f"Model 2: {self.form}")
        out = self.fitted.f_test(hypothesis)
        pvalue = ifelse(out.pvalue < 0.001, "< .001", round(out.pvalue, dec))
        print(
            f"F-statistic: {round(out.fvalue, dec)} df ({out.df_num:,.0f}, {out.df_denom:,.0f}), p.value {pvalue}"
        )



from typing import Union, Optional
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import statsmodels as sm
from statsmodels.genmod.families import Binomial
from statsmodels.genmod.families.links import logit
import statsmodels.formula.api as smf
from statsmodels.stats.outliers_influence import variance_inflation_factor
from scipy import stats
from scipy.special import expit
from .utils import ifelse
from .stats import weighted_mean, weighted_sd
from .perf import auc
from .model import sig_stars, model_fit, or_ci, or_plot, sim_prediction

# from statsmodels.regression.linear_model import RegressionResults as rrs
from .visualize import pred_plot_sm, vimp_plot_sm, extract_evars, extract_rvar


class logistic:
    def __init__(
        self,
        dataset: pd.DataFrame,
        rvar: Optional[str] = None,
        lev: Optional[str] = None,
        evar: Optional[list[str]] = None,
        form: Optional[str] = None,
    ) -> None:
        """
        Initialize logistic regression model

        Parameters
        ----------
        dataset: pandas DataFrame; dataset
        evar: List of strings; contains the names of the columns of data to be used as explanatory variables
        lev: String; name of the level in the response variable
        rvar: String; name of the column to be used as the response variable
        form: String; formula for the regression equation to use if evar and rvar are not provided
        """
        self.dataset = dataset
        self.rvar = rvar
        self.lev = lev
        self.evar = evar
        self.form = form

        if self.form:
            self.fitted = smf.glm(
                formula=self.form, data=self.dataset, family=Binomial(link=logit())
            ).fit()
            self.evar = extract_evars(self.fitted.model, self.dataset.columns)
            self.rvar = extract_rvar(self.fitted.model, self.dataset.columns)
            self.lev = self.dataset.at[0, self.rvar]
        else:
            self.form = f"{self.rvar} ~ {' + '.join(self.evar)}"
            self.fitted = smf.glm(
                formula=self.form, data=self.dataset, family=Binomial(link=logit())
            ).fit()

        df = pd.DataFrame(np.exp(self.fitted.params), columns=["OR"]).dropna()
        df["OR%"] = 100 * ifelse(df["OR"] < 1, -(1 - df["OR"]), df["OR"] - 1)
        df["coefficient"] = self.fitted.params
        df["std.error"] = self.fitted.params / self.fitted.tvalues
        # wierd but this is what statsmodels uses in summary
        df["z.value"] = self.fitted.tvalues
        df["p.value"] = self.fitted.pvalues
        df["  "] = sig_stars(self.fitted.pvalues)
        self.coef = df.reset_index()

    def summary(self, dec=3) -> None:
        """
        Summarize output from a logistic regression model
        """

        if hasattr(self.dataset, "description"):
            data_name = self.dataset.description.split("\n")[0].split()[1].lower()
        else:
            data_name = "Not available"

        print("Logistic regression (GLM)")
        print(f"Data                 : {data_name}")
        print(f"Response variable    : {self.rvar}")
        print(f"Level                : {self.lev}")
        print(f"Explanatory variables: {', '.join(self.evar)}")
        print(f"Null hyp.: there is no effect x on {self.rvar}")
        print(f"Alt. hyp.: there is an effect of x on {self.rvar}")

        df = self.coef.copy()
        df["OR"] = df["OR"].round(dec)
        df["coefficient"] = df["coefficient"].round(2)
        df["std.error"] = df["std.error"].round(dec)
        df["z.value"] = df["z.value"].round(dec)
        df["p.value"] = ifelse(
            df["p.value"] < 0.001, "< .001", df["p.value"].round(dec)
        )
        df["OR%"] = [f"{round(o, max(dec-2, 0))}%" for o in df["OR%"]]
        print(f"\n{df.to_string(index=False)}")
        print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")
        print(f"\n{model_fit(self.fitted)}")

    def ci(self, alpha=0.05, intercept=False, dec=3) -> None:
        """
        Confidence intervals for Odds Ratios
        """
        print(
            f"\n{or_ci(self.fitted, alpha=alpha, intercept=intercept).to_string(index=False)}"
        )

    def plot(
        self,
        plots="or",
        alpha=0.05,
        intercept=False,
        incl=None,
        excl=None,
        incl_int=[],
        fix=True,
        hline=False,
        figsize=None,
    ) -> None:
        """
        Plots for a logistic regression model
        """
        if "or" in plots:
            or_plot(
                self.fitted,
                alpha=alpha,
                intercept=intercept,
                incl=incl,
                excl=excl,
                figsize=figsize,
            )
        if "pred" in plots:
            pred_plot_sm(
                self.fitted,
                self.dataset,
                incl=incl,
                excl=[],
                incl_int=incl_int,
                fix=fix,
                hline=hline,
                nnv=20,
                minq=0.025,
                maxq=0.975,
            )
        if "vimp" in plots:
            vimp_plot_sm(self.fitted, self.dataset, rep=10, ax=None, ret=False)



import numpy as np
from pyrsm.bins import xtile, bincode


def test_xtile():
    x = np.array(range(10))
    bins = xtile(x, 5)
    assert all(
        bins == np.array([1, 1, 2, 2, 3, 3, 4, 4, 5, 5])
    ), "Incorrect bins returned"


def test_xtile_rev():
    x = np.array(range(10))
    bins = xtile(x, 5, rev=True)
    assert all(
        bins == np.array([5, 5, 4, 4, 3, 3, 2, 2, 1, 1])
    ), "Incorrect reversed bins returned"


def test_xtile_nan():
    x = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, np.NaN])
    bins = xtile(x, 5)
    assert all(
        bins[:9] == np.array([1, 1, 2, 2, 3, 4, 4, 5, 5])
    ), "Incorrect bins with NaN returned"
    assert np.isnan(bins[-1]), "No missing value returned"


def test_bincode_nan():
    x = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, np.NaN])
    breaks = np.quantile(x[np.isnan(x) == False], np.array(range(0, 6)) / 5)
    bins = bincode(x, breaks)
    assert all(
        bins[:9] == np.array([1, 1, 2, 2, 3, 4, 4, 5, 5])
    ), "Incorrect bins with NaN returned"
    assert np.isnan(bins[-1]), "No missing value returned"



from cmath import sqrt
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats
import seaborn as sns
from .model import sig_stars
from .utils import ifelse
from typing import Any, Optional
from scipy.stats import chisquare


class cross_tabs:
    def __init__(self, df: pd.DataFrame, var1: str, var2: str) -> None:
        """
        Calculate a Chi-square test between two categorical variables contained
        in a Pandas dataframe

        Parameters
        ----------
        df : Pandas dataframe with numeric variables

        Returns
        -------
        Cross object with several attributes
        df: Original dataframe
        var1: Name of the first categorical variable
        var2: Name of the second categorical variable
        observed: Dataframe of observed frequencies
        expected: Dataframe of expected frequencies
        expected_low: List with number of cells with expected values < 5
            and the total number of cells
        chisq: Dataframe of chi-square values for each cell
        dev_std: Dataframe of standardized deviations from the expected table
        perc_row: Dataframe of observation percentages conditioned by row
        perc_col: Dataframe of observation percentages conditioned by column
        perc: Dataframe of observation percentages by the total number of observations

        Examples
        --------
        import pyrsm as rsm
        rsm.load_data(pkg="basics", name="newspaper", dct=globals())
        ct = rsm.cross_tabs(newspaper, "Income", "Newspaper")
        ct.expected
        """
        self.df = df
        self.var1 = var1
        self.var2 = var2

        self.observed = pd.crosstab(
            df[var1], columns=df[var2], margins=True, margins_name="Total"
        )
        self.chisq_test = stats.chi2_contingency(
            self.observed.drop(columns="Total").drop("Total", axis=0), correction=False
        )
        expected = pd.DataFrame(self.chisq_test[3])
        self.expected_low = [
            (expected < 5).sum().sum(),
            expected.shape[0] * expected.shape[0],
        ]
        expected["Total"] = expected.sum(axis=1)
        expected = pd.concat(
            [expected, expected.sum().to_frame().T], ignore_index=True, axis=0
        ).set_index(self.observed.index)
        expected.columns = self.observed.columns
        self.expected = expected
        chisq = (self.observed - self.expected) ** 2 / self.expected
        chisq["Total"] = chisq.sum(axis=1)
        chisq.loc["Total", :] = chisq.sum()
        self.chisq = chisq
        self.dev_std = (
            ((self.observed - self.expected) / np.sqrt(self.expected))
            .drop(columns="Total")
            .drop("Total", axis=0)
        )
        self.perc_row = self.observed.div(self.observed["Total"], axis=0)
        self.perc_col = self.observed.div(self.observed.loc["Total", :], axis=1)
        self.perc = self.observed / self.observed.loc["Total", "Total"]

    def summary(
        self, output: list[str] = ["observed", "expected"], dec: int = 2
    ) -> None:
        """
        Print different output tables for a cross_tabs object

        Parameters
        ----------
        output : list of tables to show
            Options include "observed" (observed frequencies),
            "expected" (expected frequencies), "chisq" (chi-square values)
            for each cell, "dev_std" (standardized deviations from expected)
            "perc_row" (percentages conditioned by row), "perc_col"
            (percentages conditioned by column), "perc" (percentages by the
            total number of observations). The default value is ["observed", "expected"]
        dec : int
            Number of decimal places to use in rounding

        Examples
        --------
        import pyrsm as rsm
        rsm.load_data(pkg="basics", name="newspaper", dct=globals())
        ct = rsm.cross_tabs(newspaper)
        ct.summary()
        """

        output = ifelse(isinstance(output, str), [output], output)
        prn = f"""
Cross-tabs
Variables: {self.var1}, {self.var2}
Null hyp: there is no association between {self.var1} and {self.var2}
Alt. hyp: there is an association between {self.var1} and {self.var2}
"""
        if "observed" in output:
            prn += f"""
Observed:

{self.observed.applymap(lambda x: "{:,}".format(x))}
"""
        if "expected" in output:
            prn += f"""
Expected: (row total x column total) / total

{self.expected.round(dec).applymap(lambda x: "{:,}".format(x))}
"""
        if "chisq" in output:
            prn += f"""
Contribution to chi-squared: (o - e)^2 / e

{self.chisq.round(dec).applymap(lambda x: "{:,}".format(x))}
"""

        if "dev_std" in output:
            prn += f"""
Deviation standardized: (o - e) / sqrt(e)

{self.dev_std.round(dec).applymap(lambda x: "{:,}".format(x))}
"""

        if "perc_row" in output:
            prn += f"""
Row percentages:

{self.perc_row.transform(lambda x: (100*x).round(dec).astype(str) + "%")}
"""
        if "perc_col" in output:
            prn += f"""
Column percentages:

{self.perc_col.transform(lambda x: (100*x).round(dec).astype(str) + "%")}
"""
        if "perc_all" in output:
            prn += f"""
Percentages:

{self.perc.transform(lambda x: (100*x).round(dec).astype(str) + "%")}
"""
        prn += f"""
Chi-squared: {round(self.chisq_test[0], dec)} df({int(self.chisq_test[2])}), p.value {ifelse(self.chisq_test[1] < 0.001, "< .001", round(self.chisq_test[1], dec))}
{100 * round(self.expected_low[0] / self.expected_low[1], dec)}% of cells have expected values below 5
"""
        print(prn)

    def plot(self, output: list[str] = "perc_col", **kwargs) -> None:
        """
        Plot of correlations between numeric variables in a Pandas dataframe

        Parameters
        ----------
        output : list of tables to show
            Options include "observed" (observed frequencies),
            "expected" (expected frequencies), "chisq" (chi-square values)
            for each cell, "dev_std" (standardized deviations from expected)
            "perc_row" (percentages conditioned by row), "perc_col"
            (percentages conditioned by column), "perc" (percentages by the
            total number of observations). The default value is ["observed", "expected"]
        **kwargs : Named arguments to be passed to pandas plotting functions

        Examples
        --------
        import pyrsm as rsm
        rsm.load_data(pkg="basics", name="newspaper", dct=globals())
        ct = rsm.cross_tabs(newspaper, "Income", "Newspaper")
        ct.plot()
        """
        output = ifelse(isinstance(output, str), [output], output)

        args = {"rot": False}
        if "observed" in output:
            df = (
                self.observed.transpose()
                .drop(columns="Total")
                .drop("Total", axis=0)
                .apply(lambda x: x * 100 / sum(x), axis=1)
            )
            args["title"] = "Observed frequencies"
            args.update(**kwargs)
            fig = df.plot.bar(stacked=True, **args)
        if "expected" in output:
            df = (
                self.expected.transpose()
                .drop(columns="Total")
                .drop("Total", axis=0)
                .apply(lambda x: x * 100 / sum(x), axis=1)
            )
            args["title"] = "Expected frequencies"
            args.update(**kwargs)
            fig = df.plot.bar(stacked=True, **args)
        if "chisq" in output:
            df = self.chisq.transpose().drop(columns="Total").drop("Total", axis=0)
            args["title"] = "Contribution to chi-squared statistic"
            args.update(**kwargs)
            fig = df.plot.bar(**args)
        if "dev_std" in output:
            df = self.dev_std.transpose()
            args["title"] = "Deviation standardized"
            args.update(**kwargs)
            fig, ax = plt.subplots()
            df.plot.bar(**args, ax=ax)
            ax.axhline(y=1.96, color="black", linestyle="--")
            ax.axhline(y=1.64, color="black", linestyle="--")
            ax.axhline(y=-1.96, color="black", linestyle="--")
            ax.axhline(y=-1.64, color="black", linestyle="--")
            ax.annotate("95%", xy=(0, 2.1), va="bottom", ha="center")
            ax.annotate("90%", xy=(0, 1.4), va="top", ha="center")
        if "perc_col" in output:
            df = self.perc_col.transpose().drop(columns="Total").drop("Total", axis=0)
            args["title"] = "Column percentages"
            args.update(**kwargs)
            fig = df.plot.bar(**args)
        if "perc_row" in output:
            df = self.perc_row.transpose().drop(columns="Total").drop("Total", axis=0)
            args["title"] = "Row percentages"
            args.update(**kwargs)
            fig = df.plot.bar(**args)
        if "perc" in output:
            df = self.perc.transpose().drop(columns="Total").drop("Total", axis=0)
            args["title"] = "Table percentages"
            args.update(**kwargs)
            fig = df.plot.bar(**args)


class correlation:
    def __init__(
        self, df: pd.DataFrame, figsize: Optional[tuple[float, float]] = None
    ) -> None:
        """
        Calculate correlations between numeric variables in a Pandas dataframe

        Parameters
        ----------
        df : Pandas dataframe with numeric variables

        Returns
        -------
        Correlation object with two key attributes
        cr: Correlation matrix
        cp: p.value matrix

        Examples
        --------
        import pandas as pd
        import pyrsm as rsm
        rsm.load_data(pkg="basics", name="salary", dct=globals())
        cr = rsm.correlation(salary[["salary", "yrs.since.phd", "yrs.service"]])
        c.cr
        """
        df = df.copy()
        isNum = [
            col for col in df.columns if pd.api.types.is_numeric_dtype(df[col].dtype)
        ]
        df = df[isNum]

        ncol = df.shape[1]
        cr = np.zeros([ncol, ncol])
        cp = cr.copy()
        for i in range(ncol - 1):
            for j in range(i + 1, ncol):
                cdf = df.iloc[:, [i, j]]
                # pairwise deletion
                cdf = cdf[~np.any(np.isnan(cdf), axis=1)]
                c = stats.pearsonr(cdf.iloc[:, 0], cdf.iloc[:, 1])
                cr[j, i] = c[0]
                cp[j, i] = c[1]

        self.df = df
        self.cr = cr
        self.cp = cp
        self.figsize = figsize

    def summary(self, cutoff: float = 0, dec: int = 2) -> None:
        """
        Print correlations between numeric variables in a Pandas dataframe

        Parameters
        ----------
        cutoff : float
            Only show correlations larger than a threshold in absolute value
        dec : int
            Number of decimal places to use in rounding

        Examples
        --------
        import pandas as pd
        import pyrsm as rsm
        rsm.load_data(pkg="basics", name="salary", dct=globals())
        cr = rsm.correlation(salary[["salary", "yrs.since.phd", "yrs.service"]])
        cr.summary()
        """
        ind = np.triu_indices(self.cr.shape[0])
        cn = self.df.columns[:-1]
        indn = self.df.columns[1:]

        # correlations
        crs = self.cr.round(dec).astype(str)
        if cutoff > 0:
            crs[np.abs(self.cr) < cutoff] = ""
        crs[ind] = ""
        crs = pd.DataFrame(
            np.delete(np.delete(crs, 0, axis=0), crs.shape[1] - 1, axis=1),
            columns=cn,
            index=indn,
        )

        # pvalues
        cps = self.cp.round(dec).astype(str)
        if cutoff > 0:
            cps[np.abs(self.cr) < cutoff] = ""
        cps[ind] = ""
        cps = pd.DataFrame(
            np.delete(np.delete(cps, 0, axis=0), cps.shape[1] - 1, axis=1),
            columns=cn,
            index=indn,
        )

        cn = self.df.columns
        if len(cn) > 2:
            x = "x"
            y = "y"
        else:
            x = cn[0]
            y = cn[1]

        prn = "Correlation\n"
        prn += "Variables: " + ", ".join(list(self.df.columns)) + "\n"
        prn += f"Null hyp.: variables {x} and {y} are not correlated\n"
        prn += f"Alt. hyp.: variables {x} and {y} are correlated\n"
        print(prn)
        print("Correlation matrix:")
        print(crs)
        print("\np.values:")
        print(cps)

    def plot(self, nobs: int = 1000, dec: int = 2, figsize: tuple[float, float] = None):
        """
        Plot of correlations between numeric variables in a Pandas dataframe

        Parameters
        ----------
        nobs : int
            Number of observations to use for the scatter plots. The default
            value is 1,000. To use all observations in the plots, use nobs=-1
        dec : int
            Number of decimal places to use in rounding
        figsize : tuple
            A tuple that determines the figure size. If None, size is
            determined based on the number of numeric variables in the
            data

        Examples
        --------
        import pandas as pd
        import pyrsm as rsm
        rsm.(pkg="basics", name="salary", dct=globals())
        cr = rsm.correlation(salary[["salary", "yrs.since.phd", "yrs.service"]])
        cr.plot(figsize=(7, 3))
        """

        def cor_label(label, longest, ax_sub):
            ax_sub.axes.xaxis.set_visible(False)
            ax_sub.axes.yaxis.set_visible(False)
            # set font size to avoid exceeding boundaries
            font = (80 * self.fig_size) / (len(longest) * len(self.df.columns))
            ax_sub.text(
                0.5,
                0.5,
                label,
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=font,
            )

        def cor_text(r, p, ax_sub, dec=2):
            if np.isnan(p):
                p = 1

            p = round(p, dec)
            rt = round(r, dec)
            p_star = sig_stars([p])[0]

            font = (50 * self.fig_size) / (len(str(rt)) * len(self.df.columns))
            font_star = (12 * self.fig_size) / len(self.df.columns)

            ax_sub.axes.xaxis.set_visible(False)
            ax_sub.axes.yaxis.set_visible(False)
            ax_sub.text(
                0.5,
                0.5,
                rt,
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=font * abs(r),
            )
            ax_sub.text(
                0.8,
                0.8,
                p_star,
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=font_star,
                color="blue",
            )

        def cor_plot(x_data, y_data, ax_sub, s_size):
            sns.regplot(
                x=x_data,
                y=y_data,
                ax=ax_sub,
                scatter_kws={"alpha": 0.3, "color": "black", "s": s_size},
                line_kws={"color": "blue"},
            )

            ax_sub.axes.xaxis.set_visible(False)
            ax_sub.axes.yaxis.set_visible(False)

        def cor_mat(df, cmat, pmat, dec=2, nobs=1000, figsize=None):

            cn = df.columns
            ncol = len(cn)
            longest = max(cn, key=len)
            s_size = 50 / len(self.df.columns)

            if figsize is None:
                figsize = (max(5, cmat.shape[0]), max(cmat.shape[0], 5))

            self.fig_size = min(figsize[0], figsize[1])
            s_size = (5 * self.fig_size) / len(self.df.columns)

            _, axes = plt.subplots(ncol, ncol, figsize=figsize)

            if nobs < df.shape[0] and nobs != np.Inf and nobs != -1:
                df = df.copy().sample(nobs)

            for i in range(ncol):
                for j in range(ncol):
                    if i == j:
                        cor_label(cn[i], longest, axes[i, j])
                    elif i > j:
                        cor_plot(df[cn[i]], df[cn[j]], axes[i, j], s_size)
                    else:
                        cor_text(cmat[j, i], pmat[j, i], axes[i, j], dec=2)

            plt.subplots_adjust(wspace=0.04, hspace=0.04)
            # plt.show()

        cor_mat(self.df, self.cr, self.cp, dec=dec, nobs=nobs, figsize=figsize)


class prob_calc:
    # Probability calculator
    def __init__(self, distribution: str, params: dict) -> None:
        self.distribution = distribution
        self.__dict__.update(params)
        print("Probability calculator")

    def summary(self):
        print(f"Distribution: {self.distribution}")

        def calc_f_dist(
            df1: int, df2: int, lb: float = 0, ub: float = 0.95, decimals: int = 3
        ) -> tuple[float, float]:
            print(f"Df 1        : {df1}")
            print(f"Df 2        : {df2}")
            print(f"Mean        : {round(stats.f.mean(df1, df2, loc=lb), decimals)}")
            print(f"Variance    : {round(stats.f.var(df1, df2, loc=lb), decimals)}")
            print(f"Lower bound : {lb}")
            print(f"Upper bound : {ub}\n")

            if lb == 0:
                critical_f = round(stats.f.ppf(q=ub, dfn=df1, dfd=df2), decimals)

                _num_decimal_places_in_ub = len(str(ub).split(".")[-1])

                print(f"P(X < {critical_f}) = {ub}")
                print(
                    f"P(X > {critical_f}) = {round(1 - ub, _num_decimal_places_in_ub)}"
                )
                return (0, critical_f)

            critical_f_lower = round(stats.f.ppf(q=lb, dfn=df1, dfd=df2), decimals)

            _num_decimal_places_in_lb = len(str(lb).split(".")[-1])

            print(f"P(X < {critical_f_lower}) = {lb}")
            print(
                f"P(X > {critical_f_lower}) = {round(1 - lb, _num_decimal_places_in_lb)}"
            )
            ########################################################################################
            critical_f_upper = round(stats.f.ppf(q=ub, dfn=df1, dfd=df2), decimals)

            _num_decimal_places_in_ub = len(str(ub).split(".")[-1])

            print(f"P(X < {critical_f_upper}) = {ub}")
            print(
                f"P(X > {critical_f_upper}) = {round(1 - ub, _num_decimal_places_in_ub)}"
            )
            ########################################################################################
            _num_decimal_places = max(
                len(str(ub).split(".")[-1]), len(str(lb).split(".")[-1])
            )

            print(
                f"P({critical_f_lower} < X < {critical_f_upper}) = {round((ub - lb), _num_decimal_places)}"
            )
            print(
                f"1 - P({critical_f_lower} < X < {critical_f_upper} = {round(1 - (ub - lb), _num_decimal_places)}"
            )

            return (critical_f_lower, critical_f_upper)

        def calc_t_dist(
            df: int, lb: float = 0, ub: float = 0.95, decimals: int = 3
        ) -> tuple[float, float]:
            print(f"Df          : {df}")
            print(f"Mean        : {round(stats.t.mean(df), decimals)}")
            print(f"St. dev     : {round(stats.t.std(df), decimals)}")
            print(f"Lower bound : {lb}")
            print(f"Upper bound : {ub}\n")

            if lb == 0:
                critical_t = round(stats.t.ppf(q=ub, df=df), decimals)

                _num_decimal_places_in_ub = len(str(ub).split(".")[-1])

                print(f"P(X < {critical_t}) = {ub}")
                print(
                    f"P(X > {critical_t}) = {round(1 - ub, _num_decimal_places_in_ub)}"
                )
                return (0, critical_t)

            critical_t_lower = round(stats.t.ppf(q=lb, df=df), decimals)

            _num_decimal_places_in_lb = len(str(lb).split(".")[-1])

            print(f"P(X < {critical_t_lower}) = {lb}")
            print(
                f"P(X > {critical_t_lower}) = {round(1 - lb, _num_decimal_places_in_lb)}"
            )
            ########################################################################################
            critical_t_upper = round(stats.t.ppf(q=ub, df=df), decimals)

            _num_decimal_places_in_ub = len(str(ub).split(".")[-1])

            print(f"P(X < {critical_t_upper}) = {ub}")
            print(
                f"P(X > {critical_t_upper}) = {round(1 - ub, _num_decimal_places_in_ub)}"
            )
            ########################################################################################
            _num_decimal_places = max(
                len(str(ub).split(".")[-1]), len(str(lb).split(".")[-1])
            )

            print(
                f"P({critical_t_lower} < X < {critical_t_upper}) = {round((ub - lb), _num_decimal_places)}"
            )
            print(
                f"1 - P({critical_t_lower} < X < {critical_t_upper}) = {round(1 - (ub - lb), _num_decimal_places)}"
            )

            return (critical_t_lower, critical_t_upper)

        if self.distribution == "F":
            lb = self.lb if "lb" in self.__dict__ else 0
            ub = self.ub if "ub" in self.__dict__ else 0.95
            df1 = self.df1
            df2 = self.df2
            decimals = self.decimals if "decimals" in self.__dict__ else 3
            calc_f_dist(df1, df2, lb, ub, decimals)

        elif self.distribution == "t":
            lb = self.lb if "lb" in self.__dict__ else 0
            ub = self.ub if "ub" in self.__dict__ else 0.95
            df = self.df
            decimals = self.decimals if "decimals" in self.__dict__ else 3
            calc_t_dist(df, lb, ub, decimals)

    def plot(self):
        def plot_f_dist(
            df1: int, df2: int, lb: float = 0, ub: float = 0.95, decimals: int = 3
        ):
            x = np.linspace(stats.f.ppf(0, df1, df2), stats.f.ppf(0.99, df1, df2), 200)
            pdf = stats.f.pdf(x, df1, df2)
            plt.plot(x, pdf, "black", lw=1, alpha=0.6, label="f pdf")

            if lb == 0:
                critical_f = round(stats.f.ppf(q=ub, dfn=df1, dfd=df2), decimals)
                plt.fill_between(x, pdf, where=(x < critical_f), color="slateblue")
                plt.fill_between(x, pdf, where=(x > critical_f), color="salmon")
            else:
                critical_f_lower = round(stats.f.ppf(q=lb, dfn=df1, dfd=df2), decimals)
                critical_f_upper = round(stats.f.ppf(q=ub, dfn=df1, dfd=df2), decimals)

                plt.fill_between(
                    x,
                    pdf,
                    where=((x > critical_f_upper) | (x < critical_f_lower)),
                    color="slateblue",
                )
                plt.fill_between(
                    x,
                    pdf,
                    where=((x > critical_f_upper) | (x < critical_f_lower)),
                    color="salmon",
                )

        def plot_t_dist(
            df: int, lb: float = 0.025, ub: float = 0.975, decimals: int = 3
        ):
            x = np.linspace(stats.t.ppf(0.01, df), stats.t.ppf(0.99, df), 200)
            pdf = stats.t.pdf(x, df)
            plt.plot(x, pdf, "black", lw=1, alpha=0.6, label="t pdf")

            if lb == 0:
                critical_t = round(stats.t.ppf(q=ub, df=df), decimals)
                plt.fill_between(x, pdf, where=(x < critical_t), color="slateblue")
                plt.fill_between(x, pdf, where=(x > critical_t), color="salmon")
            else:
                critical_t_lower = round(stats.t.ppf(q=lb, df=df), decimals)
                critical_t_upper = round(stats.t.ppf(q=ub, df=df), decimals)

                plt.fill_between(
                    x,
                    pdf,
                    where=((x < critical_t_upper) | (x > critical_t_lower)),
                    color="slateblue",
                )
                plt.fill_between(
                    x,
                    pdf,
                    where=((x > critical_t_upper) | (x < critical_t_lower)),
                    color="salmon",
                )

        if self.distribution == "F":
            lb = self.lb if "lb" in self.__dict__ else 0
            ub = self.ub if "ub" in self.__dict__ else 0.95
            df1 = self.df1
            df2 = self.df2
            decimals = self.decimals if "decimals" in self.__dict__ else 3
            plot_f_dist(df1, df2, lb, ub, decimals)

        elif self.distribution == "t":
            lb = self.lb if "lb" in self.__dict__ else 0
            ub = self.ub if "ub" in self.__dict__ else 0.95
            df = self.df
            decimals = self.decimals if "decimals" in self.__dict__ else 3
            plot_t_dist(df, lb, ub, decimals)


class single_mean:
    def __init__(
        self,
        data: pd.DataFrame,
        variable: str,
        alt_hypo: str,
        conf: float,
        comparison_value: float,
    ):
        self.data = data
        self.variable = variable
        self.alt_hypo = alt_hypo
        self.conf = conf
        self.comparison_value = comparison_value

        result = stats.ttest_1samp(
            a=self.data[self.variable],
            popmean=self.comparison_value,
            nan_policy="omit",
            alternative=self.alt_hypo,
        )

        self.t_val, self.p_val = result.statistic, result.pvalue

        self.mean = np.nanmean(self.data[self.variable])
        self.n = len(self.data[self.variable])
        self.n_missing = self.data[self.variable].isna().sum()

        self.sd = self.data[self.variable].std()
        self.se = self.data[self.variable].sem()
        z_score = stats.norm.ppf((1 + self.conf) / 2)

        self.me = (z_score * self.sd / sqrt(self.n)).real
        self.diff = self.mean - self.comparison_value
        self.df = self.n - 1
        self.x_percent = self.mean - stats.t.ppf(self.conf, self.df) * self.se
        self.hundred_percent = self.mean - stats.t.ppf(0, self.df) * self.se

    def summary(self, dec=3) -> None:
        print("Single mean test")
        if hasattr(self.data, "description"):
            data_name = self.data.description.split("\n")[0].split()[1].lower()
        else:
            data_name = "Not available"
        print(f"Data: {data_name}")
        print(f"Variable: {self.variable}")
        print(f"Confidence: {self.conf}")

        print(f"Null hyp.: the mean of {self.variable} = {self.comparison_value}")

        alt_hypo = ">"
        if self.alt_hypo == "lesser":
            alt_hypo = "<"
        elif self.alt_hypo == "two-sided":
            alt_hypo = "!="

        print(
            f"Alt. hyp.: the mean of {self.variable} is {alt_hypo} {self.comparison_value}\n"
        )

        row1 = [[self.mean, self.n, self.n_missing, self.sd, self.se, self.me]]
        row2 = [
            [
                self.diff,
                self.se,
                self.t_val,
                ifelse(self.p_val < 0.001, "< .001", self.p_val),
                self.df,
                self.x_percent,
                self.hundred_percent,
            ]
        ]

        col_names1 = ["mean", "n", "n_missing", "sd", "se", "me"]
        col_names2 = [
            "diff",
            "se",
            "t.value",
            "p.value",
            "df",
            str(int((1 - self.conf) * 100)) + "%",
            "100%",
        ]

        table1 = pd.DataFrame(row1, columns=col_names1).round(dec)
        table2 = pd.DataFrame(row2, columns=col_names2).round(dec)

        print(table1.to_string(index=False))
        print(table2.to_string(index=False))

    def plot(self, types: list[str], figsize: tuple[float, float] = (10, 10)) -> None:
        numplots = 2
        which_plot = ""
        if isinstance(types, str):
            numplots = 1
            which_plot = types
        elif isinstance(types, list):
            assert len(types) == 2
            assert "hist" in types and "sim" in types
        else:
            raise TypeError("`types` should be of list data type")

        _, axes = plt.subplots(numplots, figsize=figsize)

        if numplots == 1:
            if which_plot == "hist":
                self.data[self.variable].plot.hist(
                    ax=axes, title=self.variable, color="slateblue"
                )
                plt.sca(axes)
                plt.vlines(
                    x=(self.comparison_value, self.x_percent, self.mean),
                    ymin=axes.get_ylim()[0],
                    ymax=axes.get_ylim()[1],
                    colors=("r", "k", "k"),
                    linestyles=("solid", "dashed", "solid"),
                )
            else:
                # TODO: ask Prof. Nijs about this
                pass
        else:
            self.data[self.variable].plot.hist(
                ax=axes[0], title=self.variable, color="slateblue"
            )
            plt.vlines(
                x=[self.comparison_value, self.me, self.mean],
                ymin=0,
                ymax=axes.get_ylim(),
                colors=["r", "k", "k"],
                linestyles=["solid", "dashed", "dashed"],
            )
            # simulate plot here


class compare_means:
    def __init__(
        self,
        data: pd.DataFrame,
        var1: str,
        var2: str,
        combinations: list[tuple[str, str]] = None,
        alt_hypo: str = "two-sided",
        conf: float = 0.95,
        sample_type: str = "independent",
        multiple_comp_adjustment: str = "none",
        test_type: str = "t-test",
    ) -> None:
        self.data = data
        self.var1 = var1
        self.var2 = var2
        self.combinations = combinations
        self.alt_hypo = alt_hypo
        self.conf = conf
        self.sample_type = sample_type
        self.multiple_comp_adjustment = multiple_comp_adjustment
        self.test_type = test_type

        self.t_val = None
        self.p_val = None

        def welch_dof(v1: str, v2: str) -> float:
            x = self.data[self.data[self.var1] == v1][self.var2]
            y = self.data[self.data[self.var1] == v2][self.var2]
            dof = (x.var() / x.size + y.var() / y.size) ** 2 / (
                (x.var() / x.size) ** 2 / (x.size - 1)
                + (y.var() / y.size) ** 2 / (y.size - 1)
            )

            return dof

        combinations_elements = set()
        for combination in self.combinations:
            combinations_elements.add(combination[0])
            combinations_elements.add(combination[1])
        combinations_elements = list(combinations_elements)

        rows1 = []
        mean = 0
        for element in combinations_elements:
            subset = self.data[self.data[self.var1] == element][self.var2]
            row = []
            mean = np.nanmean(subset)
            n_missing = subset.isna().sum()
            n = len(subset) - n_missing
            sd = subset.std()
            se = subset.sem()
            z_score = stats.norm.ppf((1 + self.conf) / 2)
            # was printing out imaginary part in som cases
            me = np.real(z_score * sd / sqrt(n))
            row = [element, mean, n, n_missing, sd, se, me]
            rows1.append(row)

        self.table1 = pd.DataFrame(
            rows1, columns=[self.var1, "mean", "n", "n_missing", "sd", "se", "me"]
        )

        alt_hypo_sign = " > "
        if self.alt_hypo == "less":
            alt_hypo_sign = " < "
        elif self.alt_hypo == "two-sided":
            alt_hypo_sign = " != "

        rows2 = []
        for v1, v2 in self.combinations:
            null_hypo = v1 + " = " + v2
            alt_hypo = v1 + alt_hypo_sign + v2
            diff = np.nanmean(
                self.data[self.data[self.var1] == v1][self.var2]
            ) - np.nanmean(self.data[self.data[self.var1] == v2][self.var2])

            result = stats.ttest_ind(
                self.data[self.data[self.var1] == v1][self.var2],
                self.data[self.data[self.var1] == v2][self.var2],
                equal_var=False,
                nan_policy="omit",
                alternative=self.alt_hypo,
            )

            self.t_val, self.p_val = result.statistic, result.pvalue
            se = self.data[self.data[self.var1] == v2][self.var2].sem()
            df = welch_dof(v1, v2)

            """
            Not entirely sure how to calculate these
            """
            # zero_percent = mean - stats.t.ppf(1, df) * se
            # x_percent = mean - stats.t.ppf(self.conf, df) * se

            row = [
                null_hypo,
                alt_hypo,
                diff,
                self.t_val,
                ifelse(self.p_val < 0.001, "< .001", self.p_val),
                df,
                # zero_percent,
                # x_percent,
            ]
            rows2.append(row)

        self.table2 = pd.DataFrame(
            rows2,
            columns=[
                "Null hyp.",
                "Alt. hyp.",
                "diff",
                "t.value",
                "p.value",
                "df",
                # "0%",
                # str(self.conf * 100) + "%",
            ],
        )

    def summary(self, dec=3) -> None:
        print(f"Pairwise mean comparisons {self.test_type}")
        if hasattr(self.data, "description"):
            data_name = self.data.description.split("\n")[0].split()[1].lower()
        else:
            data_name = "Not available"
        print(f"Data: {data_name}")
        print(f"Variables: {self.var1}, {self.var2}")
        print(f"Confidence: {self.conf}")
        print(f"Adjustment: {self.multiple_comp_adjustment}")

        print(self.table1.round(dec).to_string(index=False))
        print(self.table2.round(dec).to_string(index=False))


class single_prop:
    def __init__(
        self,
        data: pd.DataFrame,
        variable: str,
        level: Any,
        alt_hypo: str,
        conf: float,
        comparison_value: float,
        test_type: str = "binomial",
    ) -> None:
        self.data = data
        self.variable = variable
        self.level = level
        self.alt_hypo = alt_hypo
        self.conf = conf
        self.comparison_value = comparison_value
        self.test_type = test_type

        self.ns = len(self.data[self.data[self.variable] == self.level])
        self.n_missing = self.data[self.variable].isna().sum()
        self.n = len(self.data) - self.n_missing
        self.p = self.ns / self.n
        self.sd = sqrt(self.p * (1 - self.p)).real
        self.se = (self.sd / sqrt(self.n)).real

        z_score = stats.norm.ppf((1 + self.conf) / 2).real

        self.me = z_score * self.se

        self.diff = self.p - self.comparison_value

        results = stats.binomtest(self.ns, self.n, self.comparison_value, self.alt_hypo)

        self.p_val = results.pvalue
        proportion_ci = results.proportion_ci(self.conf)
        self.zero_percent = proportion_ci.low
        self.x_percent = proportion_ci.high

    def summary(self) -> None:
        print(f"Single proportion ({self.test_type})")
        if hasattr(self.data, "description"):
            data_name = self.data.description.split("\n")[0].split()[1].lower()
        else:
            data_name = "Not available"

        print(f"Data: {data_name}")
        print(f"Variable: {self.variable}")
        print(f"Level: {self.level} in {self.variable}")
        print(f"Confidence: {self.conf}")
        print(
            f"Null hyp.: the proportion of {self.level} in {self.variable} = {self.comparison_value}"
        )

        alt_hypo_sign = ">"
        if self.alt_hypo == "less":
            alt_hypo_sign = "<"
        elif self.alt_hypo == "two-sided":
            alt_hypo_sign = "!="

        print(
            f"Alt. hyp.: the proportion of {self.level} in {self.variable} {alt_hypo_sign} {self.comparison_value}"
        )

        row1 = [[self.p, self.ns, self.n, self.n_missing, self.sd, self.se, self.me]]
        table1 = pd.DataFrame(
            row1, columns=["p", "ns", "n", "n_missing", "sd", "se", "me"]
        )

        row2 = [[self.diff, self.ns, self.p_val, self.zero_percent, self.x_percent]]
        table2 = pd.DataFrame(
            row2, columns=["diff", "ns", "p_val", "0%", str(self.conf * 100) + "%"]
        )

        print()

        print(table1.to_string(index=False))
        print(table2.to_string(index=False))

    def plot(self) -> None:
        # TODO
        pass


class compare_props:
    def __init__(
        self,
        data: pd.DataFrame,
        grouping_var: str,
        var: str,
        level: str,
        combinations: list[tuple[str, str]],
        alt_hypo: str,
        conf: float,
        multiple_comp_adjustment: str = "none",
    ) -> None:
        self.data = data
        self.grouping_var = grouping_var
        self.var = var
        self.level = level
        self.combinations = combinations
        self.alt_hypo = alt_hypo
        self.conf = conf
        self.multiple_comp_adjustment = multiple_comp_adjustment

        # self.p_val = None

        # print("Pairwise proportion comparisons")

        # def calculate(self) -> None:
        combinations_elements = set()
        for combination in self.combinations:
            combinations_elements.add(combination[0])
            combinations_elements.add(combination[1])
        combinations_elements = list(combinations_elements)

        rows1 = []
        for element in combinations_elements:
            subset = self.data[
                (self.data[self.var] == self.level)
                & (self.data[self.grouping_var] == element)
            ][self.var]
            ns = len(subset)
            n_missing = subset.isna().sum()
            n = (
                len(self.data[(self.data[self.grouping_var] == element)][self.var])
                - n_missing
            )
            p = ns / n
            # print(f"ns: {ns}, n: {n}, p: {p}, nmissing: {n_missing}")
            sd = sqrt(p * (1 - p))
            se = sd / sqrt(n)
            z_score = stats.norm.ppf((1 + self.conf) / 2)
            # was printing out imaginary part in som cases
            me = np.real(z_score * sd / sqrt(n))
            row = [element, ns, p, n, n_missing, sd, se, me]
            rows1.append(row)

        self.table1 = pd.DataFrame(
            rows1,
            columns=[
                self.grouping_var,
                self.level,
                "p",
                "n",
                "n_missing",
                "sd",
                "se",
                "me",
            ],
        )

        alt_hypo_sign = " > "
        if self.alt_hypo == "less":
            alt_hypo_sign = " < "
        elif self.alt_hypo == "two-sided":
            alt_hypo_sign = " != "

        rows2 = []
        for v1, v2 in self.combinations:
            null_hypo = v1 + " = " + v2
            alt_hypo = v1 + alt_hypo_sign + v2

            subset1 = self.data[
                (self.data[self.var] == self.level)
                & (self.data[self.grouping_var] == v1)
            ][self.var]

            ns1 = len(subset1)
            n_missing1 = subset1.isna().sum()
            n1 = (
                len(self.data[(self.data[self.grouping_var] == v1)][self.var])
                - n_missing1
            )
            p1 = ns1 / n1

            subset2 = self.data[
                (self.data[self.var] == self.level)
                & (self.data[self.grouping_var] == v2)
            ][self.var]

            ns2 = len(subset2)
            n_missing2 = subset2.isna().sum()
            n2 = (
                len(self.data[self.data[self.grouping_var] == v2][self.var])
                - n_missing2
            )
            p2 = ns2 / n2

            diff = p1 - p2

            observed = pd.crosstab(
                self.data[v1], columns=self.data[v2], margins=True, margins_name="Total"
            )
            chisq, self.p_val, df, _ = stats.chi2_contingency(
                self.observed.drop(columns="Total").drop("Total", axis=0),
                correction=False,
            )
            # chisq, self.p_val, df, _ = stats.chi2_contingency()  # unsure about this

            # print(f"chisq: {chisq}")

            row = [
                null_hypo,
                alt_hypo,
                diff,
                self.p_val,
                chisq,
                df,
                # zero_percent,
                # x_percent,
            ]
            rows2.append(row)

        self.table2 = pd.DataFrame(
            rows2,
            columns=[
                "Null hyp.",
                "Alt. hyp.",
                "diff",
                "p.value",
                "chisq.value",
                "df",
                # "0%",
                # str(self.conf * 100) + "%",
            ],
        )

    def summary(self, dec: int = 3) -> None:
        if hasattr(self.data, "description"):
            data_name = self.data.description.split("\n")[0].split()[1].lower()
        else:
            data_name = "Not available"
        print(f"Data: {data_name}")
        print(f"Variables: {self.grouping_var}, {self.var}")
        print(f"Level: {self.level} in {self.var}")
        print(f"Confidence: {self.conf}")
        print(f"Adjustment: {self.multiple_comp_adjustment}\n")
        print(self.table1.round(dec).to_string(index=False))
        print(self.table2.round(dec).to_string(index=False))


class central_limit_theorem:
    def __init__(
        self,
        dist: str,
        sample_size: int,
        num_samples: int,
        num_bins: int,
        figsize: Optional[tuple[float, float]] = None,
        **params: float,
    ) -> None:
        self.dist = dist
        self.sample_size = sample_size
        self.num_samples = num_samples
        self.num_bins = max(1, min(num_bins, 50))
        self.figsize = figsize
        self.params = params

    def simulate(self) -> None:
        if self.dist == "normal":
            self._simulate_normal()
        elif self.dist == "binomial":
            self._simulate_binomial()
        elif self.dist == "uniform":
            self._simulate_uniform()
        elif self.dist == "exponential":
            self._simulate_exponential()
        else:
            print("Invalid distribution")

    def _simulate_normal(self) -> None:
        mean = self.params["mean"]
        sd = self.params["sd"]
        samples = [
            np.random.normal(loc=mean, scale=sd, size=self.sample_size)
            for _ in range(self.num_samples)
        ]

        self.plot_distribution(samples)

    def _simulate_binomial(self) -> None:
        size = self.params["size"]
        prob = self.params["prob"]

        samples = [
            np.random.binomial(n=size, p=prob, size=self.sample_size)
            for _ in range(self.num_samples)
        ]

        self.plot_distribution(samples)

    def _simulate_uniform(self) -> None:
        minimum = self.params["min"]
        maximum = self.params["max"]

        samples = [
            np.random.uniform(low=minimum, high=maximum, size=self.sample_size)
            for _ in range(self.num_samples)
        ]

        self.plot_distribution(samples)

    def _simulate_exponential(self) -> None:
        rate = self.params["rate"]

        samples = [
            np.random.exponential(scale=rate, size=self.sample_size)
            for _ in range(self.num_samples)
        ]

        self.plot_distribution(samples)

    def plot_distribution(self, samples: list[np.ndarray]) -> None:
        sample_means = [np.mean(sample) for sample in samples]

        _, axes = plt.subplots(2, 2, figsize=self.figsize)
        plt.subplots_adjust(wspace=0.2, hspace=0.2)

        plt.axes(axes[0][0])
        self._plot_distribution(x=samples[0], x_label="Histogram of sample #1").plot()

        plt.axes(axes[0][1])
        self._plot_distribution(
            x=samples[-1], x_label="Histogram of sample #" + str(self.num_samples)
        ).plot()

        plt.axes(axes[1][0])
        self._plot_distribution(
            x=sample_means, x_label="Histogram of sample means"
        ).plot()

        plt.axes(axes[1][1])
        axes[1][1].set_ylabel("y")
        self._plot_distribution(
            x=sample_means, x_label="Density of sample means", density_plot=True
        ).plot()

        # plt.show()

    def _plot_distribution(
        self, x: list[np.ndarray], x_label: str, density_plot: bool = False
    ) -> matplotlib.axes.Axes:
        stat = "count"
        data = {x_label: x}

        data = pd.DataFrame(data)
        if density_plot:
            return sns.kdeplot(data=data, x=x_label, fill=True)

        return sns.histplot(
            data=data,
            x=x_label,
            stat=stat,
            bins=self.num_bins,
        )


class goodness_of_fit:
    def __init__(
        self,
        data: pd.DataFrame,
        variable: str,
        figsize: tuple[float, float] = None,
        probabilities: Optional[tuple[float, ...]] = None,
        params: Optional[tuple[str, ...]] = None,
    ) -> None:
        self.data = data
        self.variable = variable
        self.figsize = figsize
        self.probabilities = probabilities
        if params is not None:
            params = map(str.lower, params)
            self.params = tuple(params)
        else:
            self.params = params

        self._observed_frequencies = self.data[self.variable].value_counts().to_dict()
        if self.params is not None:
            self._observed_df = pd.DataFrame(
                {
                    key: [
                        item,
                    ]
                    for key, item in self._observed_frequencies.items()
                },
                columns=sorted(self._observed_frequencies.keys()),
            )
            self._observed_df["Total"] = self._observed_df[
                list(self._observed_df.columns)
            ].sum(axis=1)

            self._expected_df = pd.DataFrame(
                {
                    sorted(self._observed_frequencies.keys())[i]: [
                        self.probabilities[i] * self._observed_df.at[0, "Total"],
                    ]
                    for i in range(len(self._observed_frequencies.keys()))
                },
                columns=sorted(self._observed_frequencies.keys()),
            )
            self._expected_df["Total"] = self._expected_df[
                list(self._expected_df.columns)
            ].sum(axis=1)

            self._chisquared_df = pd.DataFrame(
                {
                    column: [
                        round(
                            (
                                (
                                    self._observed_df.at[0, column]
                                    - self._expected_df.at[0, column]
                                )
                                ** 2
                            )
                            / self._expected_df.at[0, column],
                            2,
                        ),
                    ]
                    for column in self._expected_df.columns.tolist()[:-1]
                },
                columns=self._expected_df.columns.tolist(),
            )
            self._chisquared_df["Total"] = self._chisquared_df[
                list(self._chisquared_df.columns)
            ].sum(axis=1)

            self._stdev_df = pd.DataFrame(
                {
                    column: [
                        round(
                            (
                                self._observed_df.at[0, column]
                                - self._expected_df.at[0, column]
                            )
                            / sqrt(self._expected_df.at[0, column]).real,
                            2,
                        ),
                    ]
                    for column in self._expected_df.columns.tolist()[:-1]
                },
                columns=self._expected_df.columns.tolist()[:-1],
            )

    def summary(self) -> None:
        print("Goodness of fit test")
        if hasattr(self.data, "description"):
            data_name = self.data.description.split("\n")[0].split()[1].lower()
        else:
            data_name = "Not available"

        print(f"Data: {data_name}")
        if self.variable not in self.data.columns:
            print(f"{self.variable} does not exist in chosen dataset")
            return

        print(f"Variable: {self.variable}")
        num_levels = self.data[self.variable].nunique()
        if self.probabilities is None:
            self.probabilities = [1 / num_levels] * num_levels

        if num_levels != len(self.probabilities):
            print(
                f'Number of elements in "probabilities" should match the number of levels in {self.variable} ({num_levels})'
            )
            return

        prob_sum = sum(self.probabilities)
        if prob_sum != 1:
            print(f"Probabilities do not sum to 1 ({prob_sum})")
            print(
                f"Use fractions if appropriate. Variable {self.variable} has {num_levels} unique values"
            )
            return

        print(f'Specified: {" ".join(map(str, self.probabilities))}')
        print(
            f"Null hyp.: The distribution of {self.variable} is consistent with the specified distribution"
        )
        print(
            f"Alt. hyp.: The distribution of {self.variable} is not consistent with the specified distribution"
        )

        if self.params is not None:
            if "observed" in self.params:
                print("Observed:")
                print(self._observed_df.to_string(index=False))
                print()

            if "expected" in self.params:
                print("Expected: total x p")
                print(self._expected_df.to_string(index=False))
                print()

            if "chi-squared" in self.params:
                print(
                    "Contribution to chi-squared: (observed - expected) ^ 2 / expected"
                )
                print(self._chisquared_df.to_string(index=False))
                print()

            if "deviation std" in self.params:
                print("Deviation standardized: (observed - expected) / sqrt(expected)")
                print()
                print(self._stdev_df.to_string(index=False))
                print()

        chisq, p_val = chisquare(
            [
                self._observed_frequencies[key]
                for key in sorted(self._observed_frequencies.keys())
            ],
            [
                self._expected_df.at[0, key]
                for key in sorted(self._observed_frequencies.keys())
            ],
        )
        chisq = round(chisq, 3)

        if p_val < 0.001:
            p_val = "< .001"
        print(f"Chi-squared: {chisq} df ({num_levels - 1}), p.value {p_val}")

    def plot(self) -> None:
        _, axes = plt.subplots(2, 2, figsize=self.figsize)
        plt.subplots_adjust(wspace=0.2, hspace=0.2)

        if "observed" in self.params:
            plt.axes(axes[0][0])
            observed_frequency_percentages_df = pd.DataFrame(
                {
                    "levels": self._observed_df.columns.tolist()[:-1],
                    "percentages": [
                        (
                            self._observed_df.at[0, level]
                            / self._observed_df.at[0, "Total"]
                        )
                        * 100
                        for level in self._observed_df.columns.tolist()[:-1]
                    ],
                }
            )
            sns.barplot(
                data=observed_frequency_percentages_df, x="levels", y="percentages"
            ).plot()

        if "expected" in self.params:
            plt.axes(axes[0][1])
            expected_frequency_percentages_df = pd.DataFrame(
                {
                    "levels": self._expected_df.columns.tolist()[:-1],
                    "percentages": [
                        (
                            self._expected_df.at[0, level]
                            / self._expected_df.at[0, "Total"]
                        )
                        * 100
                        for level in self._expected_df.columns.tolist()[:-1]
                    ],
                }
            )
            sns.barplot(
                data=expected_frequency_percentages_df, x="levels", y="percentages"
            ).plot()

        if "chi-squared" in self.params:
            plt.axes(axes[1][0])
            chisquared_contribution_df = pd.DataFrame(
                {
                    "levels": self._chisquared_df.columns.tolist()[:-1],
                    "contribution": [
                        self._chisquared_df.at[0, level]
                        for level in self._chisquared_df.columns.tolist()[:-1]
                    ],
                }
            )
            sns.barplot(
                data=chisquared_contribution_df, x="levels", y="contribution"
            ).plot()

        if "deviation std" in self.params:
            plt.axes(axes[1][1])
            standardized_deviation_df = pd.DataFrame(
                {
                    "levels": self._stdev_df.columns.tolist(),
                    "stdev": [
                        self._stdev_df.at[0, level]
                        for level in self._stdev_df.columns.tolist()
                    ],
                }
            )

            barplot = sns.barplot(data=standardized_deviation_df, x="levels", y="stdev")

            z_95, z_neg_95 = 1.96, -1.96
            z_90, z_neg_90 = 1.64, -1.64

            barplot.axhline(y=z_95, color="k", linestyle="dashed", linewidth=1)
            plt.annotate(
                "95%",
                xy=(0, z_95),
                xytext=(0, z_95 + 0.1),
                color="black",
                fontsize=7,
            )

            barplot.axhline(y=z_neg_95, color="k", linestyle="dashed", linewidth=1)
            plt.annotate(
                "95%",
                xy=(1, z_neg_95),
                xytext=(1, z_neg_95 - 0.35),
                color="black",
                fontsize=7,
            )

            barplot.axhline(y=z_90, color="k", linestyle="dashed", linewidth=1)
            barplot.axhline(y=z_neg_90, color="k", linestyle="dashed", linewidth=1)

            barplot.plot()

        # plt.show()



from pyrsm.basics import correlation, cross_tabs
import numpy as np
import pandas as pd

df = pd.DataFrame({"x": [0, 1, 1, 1, 0, 0, 0], "y": [2, 1, 1, 1, 2, 2, 10]})

# ct.expected.iloc[0, :].round(6) == [1.714286, 1.714286, 0.571429, 4.0000]


def test_correlation():
    c = correlation(df)
    assert c.cr[1, 0].round(3) == -0.493, "Correlations incorrect"
    df_nan = df.copy()
    df_nan.loc[4, "x"] = np.NaN
    c = correlation(df_nan)
    assert c.cr[1, 0].round(3) == -0.567, "Correlations with np.NaN incorrect"


def test_crosstab():
    ct = cross_tabs(df, "x", "y")
    assert all(
        ct.expected.iloc[0, :].round(6) == [1.714286, 1.714286, 0.571429, 4.0000]
    ), "Cross tab expected values incorrect"



from shiny import App, render, ui, reactive, Inputs, Outputs, Session

# from htmltools import TagList, tags, div, css
from faicons import icon_svg
import webbrowser, nest_asyncio, uvicorn
import io, os, signal, black
from pathlib import Path
import pyrsm as rsm
import pandas as pd
from contextlib import redirect_stdout
from datetime import datetime
from itertools import combinations
from .utils import *

# from itables import to_html_datatable as DT


## next steps
## try qgrid for interactive data table
## shown description as markdown https://shinylive.io/py/examples/#extra-packages


class model_regress:
    def __init__(self, datasets) -> None:
        self.datasets = rsm.ifelse(
            isinstance(datasets, dict), datasets, {"dataset": datasets}
        )
        self.dataset_list = list(datasets.keys())
        self.stop_code = ""

    def shiny_ui(self):
        return ui.page_fluid(
            ui.head_content(
                ui.tags.script(
                    (Path(__file__).parent / "www/returnTextAreaBinding.js").read_text()
                ),
                ui.tags.script((Path(__file__).parent / "www/copy.js").read_text()),
                ui.tags.style((Path(__file__).parent / "www/style.css").read_text()),
                # ui.tags.style("style.css"),  # not getting picked up for some reason
            ),
            ui.layout_sidebar(
                ui.panel_sidebar(
                    ui.panel_well(
                        ui.input_select("datasets", "Datasets:", self.dataset_list),
                        width=3,
                    ),
                    ui.panel_conditional(
                        "input.tabs_regress == 'Data'",
                        ui.panel_well(
                            ui.input_checkbox("show_filter", "Show data filter"),
                            ui.panel_conditional(
                                "input.show_filter == true",
                                # ui.input_text_area(
                                input_return_text_area(
                                    "data_filter",
                                    "Data Filter:",
                                    rows=2,
                                    placeholder="Provide a filter (e.g., price >  5000) and press return",
                                ),
                                # ui.input_text_area(
                                input_return_text_area(
                                    "data_arrange",
                                    "Data arrange (sort):",
                                    rows=2,
                                    placeholder="Arrange (e.g., ['color', 'price'], ascending=[True, False])) and press return",
                                ),
                                # ui.input_text_area(
                                input_return_text_area(
                                    "data_slice",
                                    "Data slice (rows):",
                                    rows=1,
                                    placeholder="e.g., 1:50 and press return",
                                ),
                            ),
                        ),
                    ),
                    ui.panel_conditional(
                        "input.tabs_regress == 'Summary'",
                        ui.panel_well(
                            ui.input_action_button(
                                "run",
                                "Estimate model",
                                icon=icon_svg("play"),
                                class_="btn-success",
                                width="100%",
                            ),
                        ),
                        ui.panel_well(
                            ui.output_ui("ui_rvar"),
                            ui.output_ui("ui_evar"),
                            ui.input_radio_buttons(
                                "show_interactions",
                                "Interactions:",
                                choices={0: "None", 2: "2-way", 3: "3-way"},
                                inline=True,
                            ),
                            ui.panel_conditional(
                                "input.show_interactions > 0",
                                ui.output_ui("ui_interactions"),
                            ),
                            width=3,
                        ),
                        ui.panel_well(
                            ui.input_checkbox_group(
                                "controls",
                                "Controls:",
                                {
                                    "ci": "Confidence intervals",
                                    "ssq": "Sum of Squares",
                                    "vif": "VIF",
                                },
                            )
                        ),
                    ),
                    ui.panel_conditional(
                        "input.tabs_regress == 'Predict'",
                        ui.panel_well(
                            ui.input_select(
                                "pred_datasets", "Prediction data:", self.dataset_list
                            ),
                            width=3,
                        ),
                    ),
                    ui.panel_conditional(
                        "input.tabs_regress == 'Plot'",
                        ui.panel_well(
                            ui.input_select(
                                id="regress_plots",
                                label="Plots",
                                selected=None,
                                choices={
                                    "None": "None",
                                    "dist": "Distribution",
                                    "corr": "Correlation",
                                    "scatter": "Scatter",
                                    "dashboard": "Dashboard",
                                    "residual": "Residual vs Explanatory",
                                    "pred": "Prediction plots",
                                    "vimp": "Permutation importance",
                                    "coef": "Coefficient plot",
                                },
                            ),
                            ui.panel_conditional(
                                "input.regress_plots == 'corr'",
                                ui.input_select(
                                    "nobs",
                                    "Number of data points plotted:",
                                    {1000: "1,000", -1: "All"},
                                ),
                            ),
                            ui.panel_conditional(
                                "input.regress_plots == 'pred'",
                                ui.output_ui("ui_incl_evar"),
                                ui.output_ui("ui_incl_interactions"),
                            ),
                            width=3,
                        ),
                    ),
                    ui.panel_well(
                        ui.input_action_button(
                            "stop",
                            "Stop app",
                            icon=icon_svg("stop"),
                            class_="btn-danger",
                            width="100%",
                            onclick="window.close();",
                        ),
                    ),
                ),
                ui.panel_main(
                    ui.navset_tab_card(
                        ui.nav(
                            "Data",
                            ui.output_ui("show_data_code"),
                            ui.output_ui("show_data"),
                            ui.output_ui("show_description"),
                        ),
                        ui.nav(
                            "Summary",
                            ui.output_ui("show_summary_code"),
                            ui.output_text_verbatim("regress_summary"),
                        ),
                        ui.nav(
                            "Predict",
                            ui.output_ui("show_predict_code"),
                            ui.output_ui("regress_predict"),
                        ),
                        ui.nav(
                            "Plot",
                            ui.output_ui("show_plot_code"),
                            ui.output_plot("regress_plot", height="800px"),
                        ),
                        id="tabs_regress",
                    )
                ),
            ),
        )

    def shiny_server(self, input: Inputs, output: Outputs, session: Session):
        def code_formatter(code):
            cmd = self.stop_code = black.format_str(code, mode=black.Mode())
            return ui.HTML(
                f"""<pre><details><summary>Code</summary>{cmd}</details></pre>"""
            )
            # needs syntax highlighting but ui.markdown doesn't provide that yet
            # return ui.markdown(f"""```python\n{cmd}\n```""")
            # cmd = Markdown(f"\nGenerated code:\n\n```python\n{cmd}\n```")

        @reactive.Calc
        def get_data():
            data_name = input.datasets()
            data = self.datasets[data_name]
            if input.show_filter():
                if not is_empty(input.data_filter()):
                    data = data.query(input.data_filter())
                if not is_empty(input.data_arrange()):
                    data = eval(f"""data.sort_values({input.data_arrange()})""")
                if not is_empty(input.data_slice()):
                    data = eval(f"""data.iloc[{input.data_slice()}, :]""")

            types = {c: [data[c].dtype, data[c].nunique()] for c in data.columns}
            isNum = {
                c: f"{c} ({t[0].name})"
                for c, t in types.items()
                if pd.api.types.is_numeric_dtype(t[0])
            }
            isBin = {c: f"{c} ({t[0].name})" for c, t in types.items() if t[1] == 2}
            isCat = {
                c: f"{c} ({t[0].name})"
                for c, t in types.items()
                if c in isBin or pd.api.types.is_categorical_dtype(t[0]) or t[1] < 10
            }
            var_types = {
                "all": {c: f"{c} ({t[0].name})" for c, t in types.items()},
                "isNum": isNum,
                "isBin": isBin,
                "isCat": isCat,
            }

            return (
                data,
                data_name,
                var_types,
                f"# {data_name} = pd.read_pickle('{data_name}.pkl')",
            )

        @output(id="show_data_code")
        @render.ui
        def show_data_code():
            fname, _, code = get_data()[1:]
            if input.show_filter():
                code += f"""\n{fname} = {fname}"""
                if not is_empty(input.data_filter()):
                    code += f""".query("{input.data_filter()}")"""
                if not is_empty(input.data_arrange()):
                    code += f""".sort_values({input.data_arrange()})"""
                if not is_empty(input.data_slice()):
                    code += f""".iloc[{input.data_slice()}, :]"""
            return code_formatter(code)

        @output(id="show_data")
        @render.ui
        def show_data():
            data = get_data()[0]
            return (
                ui.HTML(
                    data.head(10).to_html(
                        classes="table table-striped data_preview", index=False
                    ),
                ),
                ui.p(f"Showing 10 rows out of {data.shape[0]:,}"),
            )

        @output(id="show_description")
        @render.ui
        def show_description():
            data = get_data()[0]
            if hasattr(data, "description"):
                return ui.markdown(get_data()[0].description)
            else:
                return ui.h3("No data description available")

        @output(id="ui_rvar")
        @render.ui
        def ui_rvar():
            isNum = get_data()[2]["isNum"]
            return ui.input_select(
                id="rvar",
                label="Response Variable",
                selected=None,
                choices=isNum,
            )

        @output(id="ui_evar")
        @render.ui
        def ui_evar():
            vars = get_data()[2]["all"]
            if (input.rvar() is not None) and (input.rvar() in vars):
                del vars[input.rvar()]

            return ui.input_select(
                id="evar",
                label="Explanatory Variables",
                selected=None,
                choices=vars,
                multiple=True,
                size=min(8, len(vars)),
                selectize=False,
            )

        @output(id="ui_interactions")
        @render.ui
        def ui_interactions():
            if len(input.evar()) > 1:
                choices = []
                nway = int(input.show_interactions())
                isNum = rsm.intersect(
                    list(input.evar()), list(get_data()[2]["isNum"].keys())
                )
                if len(isNum) > 0:
                    choices += qterms(isNum, nway=int(input.show_interactions()[0]))
                choices += iterms(input.evar(), nway=nway)
                return ui.input_select(
                    id="interactions",
                    label="Interactions",
                    selected=None,
                    choices=choices,
                    multiple=True,
                    size=min(8, len(choices)),
                    selectize=False,
                )

        @output(id="ui_incl_evar")
        @render.ui
        def ui_incl_evar():
            if len(input.evar()) > 1:
                return ui.input_select(
                    id="incl_evar",
                    label="Explanatory variables to include",
                    selected=None,
                    choices=input.evar(),
                    multiple=True,
                    size=min(8, len(input.evar())),
                    selectize=False,
                )

        @output(id="ui_incl_interactions")
        @render.ui
        def ui_incl_interactions():
            if len(input.evar()) > 1:
                nway = int(input.show_interactions())
                choices = iterms(input.evar(), nway=nway)
                return ui.input_select(
                    id="incl_interactions",
                    label="Interactions to include",
                    selected=None,
                    choices=choices,
                    multiple=True,
                    size=min(8, len(choices)),
                    selectize=False,
                )

        def model_code():
            rvar = input.rvar()
            evar = list(input.evar())
            fname, _, code = get_data()[1:]
            if int(input.show_interactions()) > 0 and len(input.interactions()) > 0:
                return f"""{code}\nreg = rsm.regress(dataset={fname}, rvar="{rvar}", evar={evar}, int={list(input.interactions())})"""
            else:
                return f"""{code}\nreg = rsm.regress(dataset={fname}, rvar="{rvar}", evar={evar})"""

        @reactive.Calc
        @reactive.event(input.run, ignore_none=True)
        def regress():
            now = datetime.now().time().strftime("%H:%M:%S")
            print(f"Model estimated at: {now}")
            data = get_data()[0]
            if int(input.show_interactions()) > 0:
                return rsm.regress(
                    dataset=data,
                    rvar=input.rvar(),
                    evar=list(input.evar()),
                    int=list(input.interactions()),
                )
            else:
                return rsm.regress(
                    dataset=data,
                    rvar=input.rvar(),
                    evar=list(input.evar()),
                )

        def summary_code(shiny=False):
            ctrl = input.controls()
            cmd = f"""ci={"ci" in ctrl}, ssq={"ssq" in ctrl}, vif={"vif" in ctrl}"""
            if shiny:
                cmd += """, shiny=True"""
            return f"""reg.summary({cmd})"""

        @output(id="show_summary_code")
        @render.text
        def show_summary_code():
            cmd = f"""{model_code()}\n{summary_code()}"""
            return code_formatter(cmd)

        @output(id="regress_summary")
        @render.text
        def regress_summary():
            out = io.StringIO()
            with redirect_stdout(out):
                reg = regress()  # get model object into local scope
                cmd = f"""{summary_code(shiny=True)}"""
                eval(cmd)
            return out.getvalue()

        @output(id="show_predict_code")
        @render.text
        def show_predict_code():
            cmd = f"""{model_code()}\npred = reg.predict(df={input.pred_datasets()}, ci=True)"""
            return code_formatter(cmd)

        @output(id="regress_predict")
        @render.text
        def regress_predict():
            return ui.HTML(
                regress()
                .predict(df=self.datasets[input.pred_datasets()], ci=True)
                .round(3)
                .head(10)
                .to_html(classes="table table-striped data_preview", index=False)
            )

        def plot_code():
            plots = input.regress_plots()
            if plots == "pred":
                incl = list(input.incl_evar())
                incl_int = list(input.incl_interactions())
                cmd = f""", incl={incl}"""
                if len(incl_int) > 0:
                    cmd += f""", incl_int={incl_int}"""
                cmd = f"""reg.plot(plots="{plots}" {cmd})"""
            elif plots == "corr":
                cmd = f"""reg.plot(plots="{plots}", nobs={input.nobs()})"""
            else:
                cmd = f"""reg.plot(plots="{plots}")"""
            return cmd

        @output(id="show_plot_code")
        @render.text
        def show_plot_code():
            plots = input.regress_plots()
            if plots != "None":
                cmd = f"""{model_code()}\n{plot_code()}"""
                return code_formatter(cmd)

        def plot_height():
            plots = input.regress_plots()
            if plots == "pred":
                return "800px"
            else:
                return "1000px"

        @output(id="regress_plot")
        # @render.plot(height=plot_height)
        @render.plot
        def regress_plot():
            plots = input.regress_plots()
            if plots != "None":
                reg = regress()  # get model object into local scope
                cmd = f"""{plot_code()}"""
                return eval(cmd)

        @reactive.Effect
        @reactive.event(input.stop, ignore_none=True)
        async def stop_app():
            rsm.md(f"```python\n{self.stop_code}\n```")
            await session.app.stop()
            os.kill(os.getpid(), signal.SIGTERM)


## should work based on https://shinylive.io/py/examples/#static-content
## but doesn't
# www_dir = Path(__file__).parent / "www"
# app = App(app_ui, server, static_assets=www_dir)


## uncomment for development and testing
# if __file__ == "app.py":
# if Path(__file__).name == "app.py":
# rsm.load_data(pkg="data", name="diamonds", dct=globals())
# mr = model_regress({"diamonds": diamonds, "diamonds100": diamonds.sample(100)})
# app = App(mr.shiny_ui(), mr.shiny_server)


def regress(
    data_dct: dict,
    host: str = "0.0.0.0",
    port: int = 8000,
    log_level: str = "warning",
):
    """
    Launch a Shiny-for-Python app for regression analysis
    """
    mr = model_regress(data_dct)
    nest_asyncio.apply()
    webbrowser.open(f"http://{host}:{port}")
    print(f"Listening on http://{host}:{port}")
    uvicorn.run(
        App(mr.shiny_ui(), mr.shiny_server),
        host=host,
        port=port,
        log_level=log_level,
    )



from htmltools import tags, div, css
from itertools import combinations


def input_return_text_area(id, label, value="", rows=1, placeholder=""):
    classes = ["form-control", "returnTextArea"]
    area = tags.textarea(
        value,
        id=id,
        class_=" ".join(classes),
        style=css(width="100%", height=None, resize="vertical"),
        placeholder=placeholder,
        rows=rows,
        cols=None,
        autocomplete=False,
        spellcheck=False,
    )

    def shiny_input_label(id, label=None):
        cls = "control-label" + ("" if label else " shiny-label-null")
        return tags.label(label, class_=cls, id=id + "-label", for_=id)

    return div(
        shiny_input_label(id, label),
        area,
        None,
        class_="form-group shiny-input-container",
        style=css(width="100%"),
    )


def is_empty(x):
    return x is None or all(c.isspace() for c in x)


def qterms(vars, nway=2):
    return [f"I({v} ** {p})" for p in range(2, nway + 1) for v in vars]


def iterms(vars, nway=2, sep=":"):
    cvars = list(combinations(vars, 2))
    if nway > 2:
        cvars += list(combinations(vars, nway))
    return [f"{sep}".join(c) for c in cvars]
